package com.tzld.piaoquan.recommend.server.service.recall;

import com.google.common.util.concurrent.ThreadFactoryBuilder;
import com.tzld.piaoquan.recommend.server.model.Video;
import com.tzld.piaoquan.recommend.server.service.recall.strategy.*;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.collections4.CollectionUtils;
import org.springframework.beans.BeansException;
import org.springframework.context.ApplicationContext;
import org.springframework.context.ApplicationContextAware;
import org.springframework.stereotype.Service;

import javax.annotation.PostConstruct;
import java.util.*;
import java.util.concurrent.*;

/**
 * @author dyp
 */
@Service
@Slf4j
public class RecallService implements ApplicationContextAware {

    private final Map<String, RecallStrategy> strategyMap = new HashMap<>();
    private ApplicationContext applicationContext;
    private ExecutorService pool;

    @PostConstruct
    public void init() {
        // init thread pool
        pool = new ThreadPoolExecutor(8, 32,
                0L, TimeUnit.SECONDS, new LinkedBlockingQueue<>(1000),
                new ThreadFactoryBuilder().setNameFormat("video-title-%d").build(), new ThreadPoolExecutor.AbortPolicy());
        Map<String, RecallStrategy> type = applicationContext.getBeansOfType(RecallStrategy.class);
        for (Map.Entry<String, RecallStrategy> entry : type.entrySet()) {
            RecallStrategy value = entry.getValue();
            strategyMap.put(value.getClass().getSimpleName(), value);
        }
    }

    public RecallResult recall(RecallParam param) {
        List<RecallStrategy> strategies = getRecallStrategy(param);
        CountDownLatch cdl = new CountDownLatch(strategies.size());
        List<Future<List<Video>>> recallResultFutures = new ArrayList<>(strategies.size());
        for (RecallStrategy strategy : strategies) {
            Future<List<Video>> future = pool.submit(() -> {
                List<Video> result = strategy.recall(param);
                cdl.countDown();
                return result;
            });
            recallResultFutures.add(future);
        }
        try {
            cdl.await(3000, TimeUnit.MILLISECONDS);
        } catch (InterruptedException e) {
            log.error("rov_pool_recall_with_region recall error", e);
            return null;
        }


        // TODO 重构
        // merge
        return merge(recallResultFutures, param);
    }

    private RecallResult merge(List<Future<List<Video>>> recallResultFutures, RecallParam param) {
        List<List<Video>> results = new ArrayList<>();
        for (Future<List<Video>> f : recallResultFutures) {
            try {
                results.add(f.get());
            } catch (Exception e) {
                log.error("future get error ", e);
                results.add(Collections.emptyList());
            }
        }

        // TODO 重构 merge sim recall and return recall
        Set<Long> videoIds = new HashSet<>();
        List<Video> datas = new ArrayList<>();
        if (param.getAbCode().equals("60054")
                || param.getAbCode().equals("60068")
                || param.getAbCode().equals("60081")
                || param.getAbCode().equals("60084")) {
            if (results.size() >= 2) {
                List<Video> region_recall = results.get(0);
                if (CollectionUtils.isNotEmpty(region_recall)) {
                    for (Video data : region_recall) {
                        if (!videoIds.contains(data.getVideoId())) {
                            datas.add(data);
                            videoIds.add(data.getVideoId());
                        }
                    }
                }
                List<Video> simRecall = null;
                List<Video> returnRecall = null;
                if (param.getAppType().equals("18") || param.getAppType().equals("19")) {
                    simRecall = results.get(1);
                } else {
                    if (results.size() >= 4) {
                        simRecall = results.get(3);
                    }
                    if (results.size() >= 5 && !param.getAbCode().equals("60054")) {
                        returnRecall = results.get(4);
                    }
                }

                if (simRecall != null && CollectionUtils.isNotEmpty(simRecall)) {
                    for (Video data : simRecall) {
                        if (!videoIds.contains(data.getVideoId())) {
                            datas.add(data);
                            videoIds.add(data.getVideoId());
                        }
                    }
                }

                if (returnRecall != null && CollectionUtils.isNotEmpty(returnRecall)) {
                    for (Video data : returnRecall) {
                        if (!videoIds.contains(data.getVideoId())) {
                            datas.add(data);
                            videoIds.add(data.getVideoId());
                        }
                    }
                }
                if (CollectionUtils.isNotEmpty(datas)) {
                    results.set(0, datas);
                }
            }
        }


        RecallResult result = new RecallResult();
        if (param.getAppType().equals("18") || param.getAppType().equals("19")) {
            // TODO 可能没有这个code
            if (param.getAbCode().equals("30001")
                    || param.getAbCode().equals("30002")
                    || param.getAbCode().equals("80001")) {
                result.setRovPoolRecall(results.get(0));
                result.setFlowPoolRecall(results.get(1));
            } else {
                result.setRovPoolRecall(results.get(0));
            }
        } else {
            if (CollectionUtils.isNotEmpty(results.get(1))) {
                result.setQuickPool(true);
                result.setRovPoolRecall(results.get(0));
                result.setFlowPoolRecall(results.get(1));

            } else {
                result.setRovPoolRecall(results.get(0));
                result.setFlowPoolRecall(results.get(2));
            }
        }
        return result;
    }

    private List<RecallStrategy> getRecallStrategy(RecallParam param) {
        List<RecallStrategy> strategies = new ArrayList<>();
        if (param.getAppType().equals("18") || param.getAppType().equals("19")) {
            strategies.add(strategyMap.get(SpecialRegionRecallStrategy.class.getSimpleName()));
        } else if (param.getFlowPoolAbtestGroup().equals("experimental_flow_set_level")) {
            strategies.add(strategyMap.get(SpecialRegionRecallStrategy.class.getSimpleName()));
            strategies.add(strategyMap.get(QuickFlowPoolWithLevelRecallStrategy.class.getSimpleName()));
            strategies.add(strategyMap.get(FlowPoolWithLevelRecallStrategy.class.getSimpleName()));
        } else if (param.getFlowPoolAbtestGroup().equals("experimental_flow_set_level_score")) {
            strategies.add(strategyMap.get(SpecialRegionRecallStrategy.class.getSimpleName()));
            strategies.add(strategyMap.get(QuickFlowPoolWithLevelScoreRecallStrategy.class.getSimpleName()));
            strategies.add(strategyMap.get(FlowPoolWithLevelScoreRecallStrategy.class.getSimpleName()));
        } else {
            strategies.add(strategyMap.get(SpecialRegionRecallStrategy.class.getSimpleName()));
            strategies.add(strategyMap.get(QuickFlowPoolWithScoreRecallStrategy.class.getSimpleName()));
            strategies.add(strategyMap.get(FlowPoolWithScoreRecallStrategy.class.getSimpleName()));
        }

        if (param.getAbCode().equals("60054")) {
            strategies.add(strategyMap.get(SimHotVideoRecallStrategy.class.getSimpleName()));
        } else if (param.getAbCode().equals("60068")
                || param.getAbCode().equals("60081")
                || param.getAbCode().equals("60084")) {
            strategies.add(strategyMap.get(SimHotVideoRecallStrategy.class.getSimpleName()));
            strategies.add(strategyMap.get(ReturnVideoRecallStrategy.class.getSimpleName()));
        }

        return strategies;
    }

    @Override
    public void setApplicationContext(ApplicationContext applicationContext) throws BeansException {
        this.applicationContext = applicationContext;
    }
}
